; vim: sts=4 sw=8 et ai fdm=marker syntax=avrasm

;
;                 fnordlicht firmware
;
;    for additional information please
;    see http://koeln.ccc.de/prozesse/running/fnordlicht
;
; (c) by Alexander Neumann <alexander@bumpern.de>
;    and Lars Noschinski <lars@public.noschinski.de>
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License version 2 as
; published by the Free Software Foundation.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; For more information on the GPL, please go to:
; http://www.gnu.org/copyleft/gpl.html



.nolist
.include "m8def.inc"
.list
.include "fifo_handling.inc"
.include "stack_handling.inc"

;   compile settings
; =============================================================================
; .equ SERIALDEBUG = 1               ; Set this for serial debugging output
; .equ DEVELBOARD = 1                ; Set this for develboard debugging output



;   eeprom settings, upload with 'make install-eeprom'
; =============================================================================
.equ I2CADDR            =   0b0000100 ; i2c addresses, 7bit
.equ I2CGCE             =   1       ; general call enabled

.eseg
        EEPROM_I2CADDR:     .db (I2CADDR<<1)|(I2CGCE)
.dseg



;   flags and aliases {{{
; =============================================================================
; flags (in register Flags)
.equ    PWMLastPulse    = 0         ; if bit0 is set, we're in the last pwm pulse
                                    ; set by:   timer1 interrupt
                                    ; reset by: main loop
.equ    PWMNewCycle     = 1         ; if this flag is set, processing for the next pwm cycle
                                    ; should start.
                                    ; set by:   timer1 interrupt
                                    ; reset by: main loop
.equ    PWMOverflowSlot = 2         ; if set, the next pwm timeslot is an overflow slot
.equ    ScriptEnabled   = 3         ; if bit3 is set, scripts are enabled
.equ    FlagOF1Written  = 4         ; if bit4 is set in update_timeslot_table, the timeslot
                                    ; for OF1 has already been written

; PWM properties
.equ    PWMChannels     =   3
.equ    PWMLevels       =   255     ; max brightness value
.equ    PWMBurstLevels  =   16      ; burst pulses count

; PWM output channel bits
.equ    PWMChannel0     =   0
.equ    PWMChannel1     =   1
.equ    PWMChannel2     =   2

; LedStatusTable offsets
.equ    LedRecBrightness = 0
.equ    LedRecTargetVal  = 1
.equ    LedRecFadeSpeedL = 2        ; 8.8 fixed point value, byte behind point
.equ    LedRecFadeSpeedH = 3        ; byte before point
.equ    LedRecCycleCount = 4
.equ    LedRecFlags      = 5
.equ    LedRecSize       = 6

; LedRecFlags
.equ LedFlagTargetReached = 0

; PWMActionSlots offsets
.equ    ActionRecTimeSlotL  =   0
.equ    ActionRecTimeSlotH  =   1
.equ    ActionRecBitmask    =   2
.equ    ActionRecSize       =   3

; serial port
.equ CLOCK = 16000000
.equ BAUD = 19200
.equ UBRRVAL = CLOCK/(BAUD*16)-1

; serial fifos
.equ URSize = 16        ; uart receive fifo size
.equ UTSize = 16

; i2c fifos
.equ IRSize = 8         ; i2c receive fifo size
.equ ITSize = 16        ; i2c transmit fifo size

; }}}

;   fifos {{{
; =============================================================================
.dseg
        ;   fifo: UReceiveFifo
        ; ==========================
        UReceiveFifo:
            .byte   FifoControlSize + URSize

        ;   fifo: UTransmitFifo
        ; ==========================
        UTransmitFifo:
            .byte   FifoControlSize + UTSize

        ;   fifo: IReceiveFifo
        ; ==========================
        IReceiveFifo:
            .byte   FifoControlSize + IRSize

        ;   fifo: ITransmitFifo
        ; ==========================
        ITransmitFifo:
            .byte   FifoControlSize + ITSize

; }}}

.cseg

;   register definitions {{{
; =============================================================================
.def    IReceiveFifoWPointerL   =   r4  ; i2c receive fifo temporary write pointer
.def    IReceiveFifoWPointerH   =   r5
.def    IReceiveFifoCounter     =   r6
.def    l1      =   r8      ; low working registers
.def    l2      =   r9
.def    l3      =   r10
.def    PWMNextOutputValue  =   r11  ; pwm output value for next pulse (non-burst mode)
.def    PWMBitmask  = r12   ; pwm next timeslot's bitmask
.def    w1      =   r16     ; working registers
.def    w2      =   r17
.def    w3      =   r18
.def    w4      =   r19
.def    w5      =   r20
.def    w6      =   r21
.def    Flags   =   r22     ; global flags
.def    t1w1    =  r24      ; timer1 interrupt handler working registers
.def    t1w2    =  r25

; registers Y and Z may be used freely (although Z is used in the timer1 interrupt)

; }}}

;   interrupt table {{{
; =============================================================================
.ORG 0x0                            ; reset vector
        rjmp reset

.ORG OC1Aaddr                       ; timer1 CompareA interrupt handler
        rjmp OC1AHandler
.ORG URXCaddr                       ; USART receive complete interrupt handler
        rjmp URXCHandler
.ORG UDREaddr                       ; USART data register empty handler
        rjmp UDREHandler
.ORG TWIaddr                        ; TWI data handler
        rjmp TWIHandler
.ORG 0x13                           ; program start

; }}}

;   SERIALDEBUG output routines {{{
; =============================================================================

;   DEBUG_write_uart_transmit_fifo {{{
; =============================================================================
; saves registers and calls write_uart_transmit_fifo
DEBUG_write_uart_transmit_fifo:
        push w2
        push l1
        push l2
        rcall write_uart_transmit_fifo
        pop l2
        pop l1
        pop w2
        ret
; }}}

; writes an immediate value
.macro DEBUG_WRITE_I
.ifdef SERIALDEBUG
    push w1
    ldi w1, @0
    rcall DEBUG_write_uart_transmit_fifo
    pop w1
.endif
.endmacro

; writes an register value
.macro DEBUG_WRITE_R
.ifdef SERIALDEBUG
    push w1
    mov w1, @0
    rcall DEBUG_write_uart_transmit_fifo
    pop w1
.endif
.endmacro

; }}}

;   DEVELBOARD output routines {{{
; =============================================================================

; toggles a led
.macro DEBUG_LED_TOGGLE
.ifdef DEVELBOARD
    push w1
    push w2
    in w1, PORTD
    ldi w2, (1<<@0)<<2
    eor w1, w2
    out PORTD, w1
    pop w2
    pop w1
.endif
.endmacro

; switch on a led
.macro DEBUG_LED_ON
.ifdef DEVELBOARD
    cbi PORTD, @0+2
.endif
.endmacro

; switch off a led
.macro DEBUG_LED_OFF
.ifdef DEVELBOARD
    sbi PORTD, @0+2
.endif
.endmacro


; }}}

; We need the register definitions before including this file
.include "script_handler.inc"
.include "remote_exec_handler.inc"

;   interrupt handler
; =============================================================================

;   function: reset interrupt handler {{{
; =============================================================================
reset:

        ;  init stack
        ; ========================================
        ldi w1, low(RAMEND)
        out SPL, w1
        ldi w1, high(RAMEND)
        out SPH, w1

        ;  init outputs and set default output
        ; ========================================
        ldi w1, (1<<PB0) | (1<<PB1) | (1<<PB2)   ; configure PB0-PB2 as output
        out DDRB, w1
        ldi w1, 0b111               ; all channels high, leds off
        out PORTB, w1

        .ifdef DEVELBOARD
            ldi w1, (1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7) ; configure PD2-PD7 as output
            out DDRD, w1
            ldi w1, 0xff                    ; all channels high, leds off
            out PORTD, w1
        .endif

        ;  init Y pointer
        ; ========================================
        ldi YH, HIGH(LedStatusTable)   ; load LedStatusTable address into Y and registers (for use with ld)
        ldi YL, LOW(LedStatusTable)

        ;  init LedStatusTable
        ; ========================================
        ldi w1, PWMChannels         ; load channel count
    resetInitLST:
        ldi w2, 0                   ; current brightness
        st Y+, w2
        ldi w2, 0                   ; target brightness
        st Y+, w2
        ldi w2, 64                  ; LedRecFadeSpeedL
        st Y+, w2
        ldi w2, 0                   ; LedRecFadeSpeedH
        st Y+, w2
        ldi w2, 0                   ; cycle counter
        st Y+, w2
        ldi w2, 0                   ; flags
        st Y+, w2
        dec w1
        brne resetInitLST           ; loop until everything is initialized

        ; FIXME
        ;ldi w1, 198
        ;sts LedStatusTable+LedRecSize*0+LedRecTargetVal, w1
        ;sts LedStatusTable+LedRecSize*0+LedRecBrightness, w1
        ;ldi w1, 213
        ;sts LedStatusTable+LedRecSize*1+LedRecTargetVal, w1
        ;sts LedStatusTable+LedRecSize*1+LedRecBrightness, w1
        ;ldi w1, 180
        ;sts LedStatusTable+LedRecSize*2+LedRecTargetVal, w1
        ;sts LedStatusTable+LedRecSize*2+LedRecBrightness, w1

        ;  init timer1
        ; ========================================
        ldi w1, high(64000)         ; initial delay before pwm loop starts, SIC!
        out OCR1AH, w1
        ldi w1, low(64000)
        out OCR1AL, w1

        ldi w1, (1<<CS10)           ; no prescaler, normal mode
        out TCCR1B, w1
        clr w1                      ; nothing else
        out TCCR1A, w1

        ;  init uart
        ; ========================================
        ldi w1, high(UBRRVAL)
        out UBRRH, w1
        ldi w1, low(UBRRVAL)
        out UBRRL, w1
        ldi w1, (1<<URSEL)|(3<<UCSZ0) ; frame format: 8N1
        out UCSRC, w1

        ;  init i2c
        ; ========================================
        ldi w1, high(EEPROM_I2CADDR)    ; load i2c address in eeprom into EEAR[LH]
        out EEARH, w1
        ldi w1, low(EEPROM_I2CADDR)
        out EEARL, w1

        sbi EECR, EERE              ; start read

        in w1, EEDR                 ; copy read data

        cpi w1, 0xff                ; check for invalid i2c address
        brne i2c_continue_init      ; (when eeprom is not programmed)

        ldi w1, 0b110               ; toggle red leds on
        out PORTB, w1

    i2c_invalid_address:            ; refuse to do anything else
        rjmp i2c_invalid_address

    i2c_continue_init:
        out TWAR, w1                ; set i2c address
        ldi w1, (1<<TWEN)|(1<<TWINT)|(1<<TWIE)|(1<<TWEA)
        out TWCR, w1

        ;  init fifos
        ; ========================================
        MACRO_FIFO_INIT UReceiveFifo
        MACRO_FIFO_INIT UTransmitFifo
        MACRO_FIFO_INIT IReceiveFifo
        MACRO_FIFO_INIT ITransmitFifo


        ; init ScriptStatusTable
        ; ========================================
        ; TODO: script initialization
        ldi w1, ScriptHandlerFlash

        ;ldi w1, ScriptHandlerDisabled ; FIXME
        sts ScriptInfoTable+ScrRecHandler, w1
        sts ScriptInfoTable+ScrRecSize+ScrRecHandler, w1
        sts ScriptInfoTable+2*ScrRecSize+ScrRecHandler, w1

        ldi w1, low(static_script_r*2)
        sts ScriptInfoTable+ScrRecPos, w1
        ldi w1, high(static_script_r*2)
        sts ScriptInfoTable+ScrRecPos+1, w1

        ldi w1, low(static_script_g*2)
        sts ScriptInfoTable+ScrRecSize+ScrRecPos, w1
        ldi w1, high(static_script_g*2)
        sts ScriptInfoTable+ScrRecSize+ScrRecPos+1, w1

        ldi w1, low(static_script_b*2)
        sts ScriptInfoTable+2*ScrRecSize+ScrRecPos, w1
        ldi w1, high(static_script_b*2)
        sts ScriptInfoTable+2*ScrRecSize+ScrRecPos+1, w1

        ldi w1, 0
        sts ScriptInfoTable+ScrRecStackOffset, w1
        sts ScriptInfoTable+ScrRecSize+ScrRecStackOffset, w1
        sts ScriptInfoTable+2*ScrRecSize+ScrRecStackOffset, w1

        sts ScriptInfoTable+ScrRecFlags, w1
        sts ScriptInfoTable+ScrRecSize+ScrRecFlags, w1
        sts ScriptInfoTable+2*ScrRecSize+ScrRecFlags, w1

        ;  enable timer1 interrupt
        ; ========================================
        ldi w1, (1<<OCIE1A)         ; enable timer1 compare interrupt
        out TIMSK, w1

        ;  enable uart interrupt
        ; ========================================
        ldi w1, (1<<RXCIE)|(1<<RXEN)|(1<<TXEN); receiver interrupt complete enable, receiver enable,
                                        ; transmitter enable
        out UCSRB, w1

        ;  init flag register
        ; ========================================
        ldi Flags, 0

        ;  init led registers and pwm burst table
        ; ========================================
        ;rcall update_led_brightness  ; update led brightness registers
        rcall update_pwm_burst_table ; fill pwm burst output table
        rcall update_timeslot_table ; create initial timeslot table

        ;  enable interrupts
        ; ========================================
        sei

        ;  output system bootup message to uart
        ; ========================================
        rcall output_string
        .db "B", 13, 10, 0

        ;  initiate script execution
        ; ========================================
        sbr Flags, (1<<ScriptEnabled)

        ;  start main procedure
        ; ========================================
        rjmp main

; }}}

;   function: timer1 interrupt handler {{{
;
; =============================================================================

    ;   macro: macro_output_next_value (3) {{{
    ; =============================================================================
    ; duration: 3 cycles
    ; parameters: @0: used register
    ; uses: @0, X
    .macro macro_output_next_value      ; load next output value from PWMBurstOutputTable and write to output    3
            ld @0, X+
            out PORTB, @0
    .endmacro

    ; }}}

    ;   macro: macro_delay (3n) {{{
    ; =============================================================================
    ; duration: 3*@0 cycles, constraints: @0 > 0
    ; parameters: @0: n
    .macro macro_delay                  ; do nothing for 3*n cycles
            ldi t1w1, @0                ; load given value                              1
    macro_delay_loop:
            dec t1w1                    ; decrement given value                         1
            brne macro_delay_loop       ; branch if nonzero                             1 if false, 2 if true
    .endmacro

    ; }}}

    ;   macro: macro_decrement_timer1_counter_register (6) {{{
    ; =============================================================================
    ; duration: 6 cycles
    ; parameters: none
    ; uses: t1w1, t1w2
    .macro macro_decrement_timer1_counter_register ; decrement the timer1 counter register by 64000
        in t1w1, TCNT1L             ; load counter                  1
        in t1w2, TCNT1H             ;                               1
        subi t1w1, low(64000-6)     ; subtract 64000                1
        sbci t1w2, high(64000-6)    ;                               1
        out TCNT1H, t1w2            ; store counter                 1
        out TCNT1L, t1w1            ;                               1

    .endmacro

    ; }}}

    ;   macro: macro_decrement_timeslot_counter (5) {{{
    ; =============================================================================
    ; duration: 5 cycles
    ; parameters: @0: used register
    .macro macro_decrement_timeslot_counter ; decrement timeslot counter by 1
        lds @0, PWMActionSlotCounter  ; load counter                  2
        dec @0                        ; decrement counter             1
        sts PWMActionSlotCounter, @0  ; store counter                 2

    .endmacro

    ; }}}

    ;   macro: macro_prepare_next_timeslot (24) {{{
    ; =============================================================================
    ; duration: 24 cycles
    ; parameters: @0: used pointer register (either X or Y)
    ; uses: t1w1, t1w2, @0
    .macro macro_prepare_next_timeslot  ; output top, set overflow flag

        lds @0L, PWMActionSlotPointer    ; load action slot record pointer       2
        lds @0H, PWMActionSlotPointer+1  ;                                       2

        ld t1w1, @0+                 ; load new top                              2
        ld t1w2, @0+                 ;                                           2
        ld PWMBitmask, @0+           ; load new bitmask                          2

        sts PWMActionSlotPointer+1, @0H  ; store pointer                         2
        sts PWMActionSlotPointer, @0L    ;                                       2

        ldi @0L, 0                   ;                                           1
        cp t1w1, @0L                 ; check if top is 0                         1
        cpc t1w2, @0L                ;                                           1 (makes 17 cycles so far)
        breq macro_next_timeslot_prepare_overflow   ; if overflow, jump away    1 if normal, 2 if overflow

        nop                         ;                                           1
        nop                         ;                                           1
        rjmp macro_next_timeslot_prepare_normal ;                               2

    macro_next_timeslot_prepare_overflow:
        ldi t1w1, low(64000)        ; load next interrupt timeslot              1
        ldi t1w2, high(64000)       ;                                           1

        sbr Flags, (1<<PWMOverflowSlot) ; set overflow-slot flag                1

    macro_next_timeslot_prepare_normal:
        out OCR1AH, t1w2            ; output TOP value                          1
        out OCR1AL, t1w1            ;                                           1
    .endmacro

    ; }}}

OC1AHandler:                        ;                                               8
        in t1w1, SREG               ; save the status register                      1
        push t1w1                   ;                                               2
        push XL
        push XH

        sbrs Flags, PWMOverflowSlot ; are we an overflow interrupt?
        rjmp OC1AHandler_no_overflow ; if not, jump away.

        cbr Flags, (1<<PWMOverflowSlot)  ; clear overflow flag

        tst PWMBitmask              ; check if bitmask is 0, if so (we are in the first overflow slot), do the burst (OF0)
        breq OC1AHandler_OF0        ; if zero, jump away, we are in the first overflow timeslot

        rjmp OC1AHandler_OF1        ; else (nonzero), wer are in the second overflow timeslot

    OC1AHandler_OF0: ; {{{

        ldi XL, low(PWMBurstOutputTable)
        ldi XH, high(PWMBurstOutputTable)


                                    ; first pulse, X has been initialized
                                    ; timing values:
                                    ;   2, 6, 10, 14, 18, 22, 26, 30
                                    ;   33, 37, 41, 45, 49, 53, 57, ...

        ld t1w1, X+                 ; preload pulse0 value                          2
        ld t1w2, X+                 ; preload pulse1 value                          2
            out PORTB, t1w1             ; output first value                            1
            nop                         ; wait 1 cycles                                 1
            out PORTB, t1w2             ; output second value, and wait 6 cycles        1
        push w1                     ; burn 3 cycles by pushing w1 (2) and a nop (1)
            nop
            macro_output_next_value w1  ; load next value, costs 3 cycles, makes 6 cycles :), afterwards wait 10 cycles
            macro_delay 2               ; burn 6 cycles
            nop                         ; and another one, makes 7
            macro_output_next_value w1  ; load next value, costs 3 cycles, makes 10 cycles :), afterwards wait 14 cycles
        macro_decrement_timer1_counter_register ; burns 6 cycles
        macro_decrement_timeslot_counter w1 ; burns 5 cycles
            macro_output_next_value w1  ; next 18
            macro_delay 5
            macro_output_next_value w1  ; next 22
            macro_delay 6
            nop
            macro_output_next_value w1  ; next 26
            macro_delay 6
            push YL
            push YH
            nop
            macro_output_next_value w1  ; next 30
        macro_prepare_next_timeslot Y   ; burns 24 cycles
            macro_delay 1
            macro_output_next_value w1  ; next 33
            macro_delay 8
            pop YH
            pop YL
            nop
            nop
            macro_output_next_value w1  ; next 37
            macro_delay 11
            nop
            macro_output_next_value w1  ; next 41
            macro_delay 12
            nop
            nop
            macro_output_next_value w1  ; next 45
            macro_delay 14
            macro_output_next_value w1  ; next 49
            macro_delay 15
            nop
            macro_output_next_value w1  ; next 53
            macro_delay 16              ; 48
        pop w1                      ; 2 (restore w1)
            macro_output_next_value t1w1; next 57
            macro_delay 18
            macro_output_next_value t1w1; last pwm output

        sbr Flags, (1<<PWMNewCycle) ; process new pwm burst table data

        rjmp OC1AHandler_cleanup

        ; }}}

    OC1AHandler_OF1: ; {{{

        macro_decrement_timer1_counter_register ; 6 cycles
        com PWMBitMask              ; since we are in OF1, PWMBitMask is 0xff
                                    ; so if we just use the complement, nothing
                                    ; is changed in PORTB (an that's the desired
                                    ; behaviour)
        rjmp OC1AHandler_prepare_next_slot

    ; }}}

    OC1AHandler_no_overflow:

    OC1AHandler_prepare_next_slot: ; {{{
        macro_decrement_timeslot_counter t1w1 ; decrement timeslot counter, next 'brne' uses SREG value set by 'dec' in the macro

        brne OC1AHandler_prepare_next_slot2  ; if nonzero, jump away

        sbr Flags, (1<<PWMLastPulse)    ; else set last pulse flag      1

    OC1AHandler_prepare_next_slot2:
        in t1w1, PORTB              ; read PORTB value
        or t1w1, PWMBitMask         ; set bits (turn off leds) by ORing
                                    ; with PWMBitMask (if in OF1, PWMBitMask
                                    ; is 0)
        out PORTB, t1w1             ; output given bitmask

        macro_prepare_next_timeslot X ; load new top, set PWMOverflowSlot flag
    ; }}}

    OC1AHandler_cleanup:
        pop XH
        pop XL
        pop t1w1                    ; restore the status register                   2
        out SREG, t1w1              ;                                               1
        reti


; }}}

;   function: uart receive complete interrupt handler {{{
; =============================================================================
URXCHandler:                        ;                                           8
        push w1                     ; save w1                                   2
        in w1, SREG                 ; save the status register                  1
        push w1                     ;                                           2

        MACRO_FIFO_LOAD_COUNTER UReceiveFifo, w1 ; load counter
        cpi w1, URSize              ; compare counter and fifo size             1
        breq URXDiscardData         ; if counter == fifosize, just discard the byte received  1 false/2 true

        push w2                     ; save some more registers                  2
        push YL                     ; save Y                                    2
        push YH                     ;                                           2

        in w1, UDR                  ; read data byte                            1

                                    ; store data into receive fifo
        MACRO_FIFO_STORE UReceiveFifo, w1, w2, URSize

URXCleanup:
        pop YH                      ; restore Y                                 2
        pop YL                      ;                                           2
        pop w2                      ;                                           2
        pop w1                      ; restore the status register               2
        out SREG, w1                ;                                           1
        pop w1                      ; restore w1                                2
        reti                        ;                                           4

URXDiscardData:
        in w1, UDR                  ; clear interrupt by reading the received byte
        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti

; }}}

;   function: uart data register empty handler {{{
; =============================================================================
UDREHandler:
        push w1                     ; save w1
        in w1, SREG                 ; save the status register
        push w1

        MACRO_FIFO_LOAD_COUNTER UTransmitFifo, w1   ; load uart transmit fifo counter into w1
        cpi w1, 1
        brsh UDRETransmitByte       ; if w1 >= 1, jump away
                                    ; else:
        in w1, UCSRB                ; disable data register empty interrupt
        cbr w1, (1<<UDRIE)
        out UCSRB, w1

        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti

UDRETransmitByte:
        push w2                     ; save registers
        push w3
        push YL                     ; save Y
        push YH

        MACRO_FIFO_READ UTransmitFifo, w1, UTSize
        out UDR, w1                 ; write data to uart data register

        pop YH                      ; restore Y
        pop YL
        pop w3                      ; restore registers
        pop w2
        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti
; }}}

;   function: TWI interrupt handler {{{
;
; =============================================================================
TWIHandler:
        push w1
        in w1, SREG
        push w1

        in w1, TWSR                     ; load twi status
        andi w1, 0b11111000             ; mask out prescaler bits

        cpi w1, 0x60                    ; sla+w+ack received
        breq normal_sla_w
        cpi w1, 0x68                    ; lost abitration in sla+r/w and own sla+w+ack received
        breq normal_sla_w
        cpi w1, 0x70                    ; gca+w+ack received
        breq normal_sla_w
        cpi w1, 0x78                    ; lost abitration in sla+r/w and gca+w+ack received
        breq normal_sla_w
        cpi w1, 0x80                    ; data received as sla+w, ack
        breq normal_sla_w_data_ack
        cpi w1, 0x88                    ; data received as sla+w, nack
        breq normal_sla_stop_bridge
        cpi w1, 0x90                    ; data received as gca+w, ack
        breq normal_sla_w_data_ack
        cpi w1, 0x98                    ; data received as gca+w, nack
        breq normal_sla_stop_bridge
        cpi w1, 0xA0
        breq normal_sla_stop_bridge

        DEBUG_WRITE_I 'X'

        rjmp normal_end

normal_sla_w: ; {{{
        push YL                             ; save Y
        push YH

        DEBUG_WRITE_I 'S'

        lds IReceiveFifoWPointerL, IReceiveFifo + FifoWPointerL ; load ireceivefifo write pointer
        lds IReceiveFifoWPointerH, IReceiveFifo + FifoWPointerH

        clr IReceiveFifoCounter
        MACRO_FIFO_STORE IReceiveFifo, IReceiveFifoCounter, w1, IRSize  ; store nullbyte

        pop YH                              ; restore Y
        pop YL

        rjmp normal_end
; }}}

normal_sla_stop_bridge:
        rjmp normal_sla_stop

normal_sla_w_data_ack: ; {{{
        push YL                             ; save Y
        push YH

        in w1, TWDR
        DEBUG_WRITE_I 'D'
        DEBUG_WRITE_R w1

        MACRO_FIFO_STORE IReceiveFifo, w1, w1, IRSize  ; store data byte

        inc IReceiveFifoCounter

        pop YH                              ; restore Y
        pop YL

        rjmp normal_end
; }}}

normal_sla_stop: ; {{{
        push YL                             ; save Y
        push YH

        DEBUG_WRITE_I 's'

        mov w1, IReceiveFifoCounter         ; load counter

        cpi w1, 0                           ; check if we received something
        breq normal_sla_stop_reset_fifo

        DEBUG_WRITE_I 'R'
        DEBUG_WRITE_R w1

        movw YH:YL, IReceiveFifoWPointerH:IReceiveFifoWPointerL ; else save counter into byte already on the fifo
        std Y+0, w1

        pop YH                              ; restore Y
        pop YL

        rjmp normal_end

    normal_sla_stop_reset_fifo:
        MACRO_FIFO_POP IReceiveFifo, w1, IRSize ; pop last written (null)byte

        pop YH                              ; restore Y
        pop YL

        DEBUG_WRITE_I 'r'

        rjmp normal_end

; }}}

normal_end: ; {{{
        in w1, TWCR
        ori w1, (1<<TWINT)|(1<<TWEA)
        out TWCR, w1

        pop w1
        out SREG, w1
        pop w1
        reti
; }}}

; }}}


;   other functions
; =============================================================================

;   function: main {{{
; =============================================================================
main:
        sbrc Flags, PWMLastPulse    ;  check for pwm pulse end
        rjmp main_last_pulse

        sbrc Flags, PWMNewCycle     ; process data for a new pwm cycle
        rjmp main_new_cycle


        rcall check_serial_input    ; check for serial input via uart

        rjmp main

main_last_pulse:
        rcall update_timeslot_table     ; update timeslot table

        cbr Flags, (1<<PWMLastPulse)    ; clear burst-end flag
        rjmp main

main_new_cycle:
        ;sbi UCSRB, RXCIE            ; enable uart receive complete interrupt
        ;in t1w1, TWCR               ; enable TWI interrupt
        ;ori t1w1, (1<<TWIE)
        ;out TWCR, t1w1

        ;rcall execute_remote_commands   ; remote commands

        sbrc Flags, ScriptEnabled       ; scripts
        rcall execute_scripts

        rcall update_status_table

        ;DEBUG_WRITE_I 'B'

        rcall update_pwm_burst_table

        cbr Flags, (1<<PWMNewCycle)   ; clear flag
        rjmp main

; }}}


;   function: output_string {{{
; =============================================================================
output_string:
        pop ZH              ; load string address from the stack
        pop ZL
        lsl ZL              ; shift left Z (using carry)
        rol ZH
output_next_char:
        lpm w1, Z+          ; load next byte into w1 and increment Z
        cpi w1, 0           ; if it's not the end of the string
        breq output_string_end ; jump away

        rcall write_uart_transmit_fifo ; transmit char
        rjmp output_next_char ; loop

output_string_end:
        adiw ZL, 1          ; increment Z to point to the next instruction after the string
        lsr ZH              ; shift right Z (using carry)
        ror ZL
        push ZL             ; push Z back onto the stack
        push ZH
        ret

; }}}

;   function: update_timeslot_table {{{
; =============================================================================

        ;   function: update_timeslot_test_of1 {{{
        ; =============================================================================
        update_timeslot_test_of1:
                sbrc Flags, FlagOF1Written ; if bit in local flag register is cleared, skip the ret
                ret                 ; return, if OF1 has already been written

                cpi w4, 181         ; w4 >= 181?
                brlo update_timeslot_test_of1_ret

                sbr Flags, (1<<FlagOF1Written); set bit in local flag register

                ldi w5, 0           ; store overflow counter
                st X+, w5
                st X+, w5
                ldi w5, 0xff        ; store bitmask to identify OF1
                st X+, w5

        update_timeslot_test_of1_ret:
                ret
        ; }}}

        ;   function: update_timeslot_write_timeslot_counter {{{
        ; =============================================================================
        update_timeslot_write_timeslot_counter:
                cpi w4, 16              ; test if timeslot is needed
                brsh do_timeslot_write

                ret

            do_timeslot_write:
                cpi w4, 255             ; test again if timeslot is needed
                brne do_timeslot_write_really

                ret

            do_timeslot_write_really:
                ldi ZL, low((PWMTimerTable-16)*2)  ; init Z register to point to the pwm timer table, prepared for use with lpm
                ldi ZH, high((PWMTimerTable-16)*2)

                ldi w5, 0

                lsl w4              ; shift left w4 for use with lpm
                brcc update_timeslot_write_timeslot_counter_add; if carry bit, we have to load 1 into w5
                ldi w5, 1

            update_timeslot_write_timeslot_counter_add:
                add ZL, w4          ; compute offset
                adc ZH, w5

                lpm w4, Z+          ; load counter low byte
                st X+, w4           ; store counter low byte
                lpm w4, Z+          ; load counter high byte
                st X+, w4           ; store counter high byte

                or w1, w6           ; include persistent mask
                st X+, w1           ; store mask
                mov w6, w1          ; copy mask

                ret
        ; }}}

update_timeslot_table:
        cbr Flags, (1<<FlagOF1Written) ; init flags; clear OF1 flag

        lds l1, LedStatusTable+LedRecSize*0+LedRecBrightness    ; load brightness values
        lds l2, LedStatusTable+LedRecSize*1+LedRecBrightness
        lds l3, LedStatusTable+LedRecSize*2+LedRecBrightness

        ldi w1, (1<<PWMChannel0)                                ; load bitmasks
        ldi w2, (1<<PWMChannel1)
        ldi w3, (1<<PWMChannel2)

        cp l2, l1           ; l2 >= l1
        brsh update_timeslot_table_compare2 ; if yes, jump away

        mov w4, l1          ; if no, swap l1<->l2 and w1<->w2
        mov l1, l2
        mov l2, w4

        mov w4, w1
        mov w1, w2
        mov w2, w4

    update_timeslot_table_compare2:
        cp l3, l2           ; l3 >= l2
        brsh update_timeslot_table_compare3 ; if yes, jump away

        mov w4, l2          ; if no, swap l2<->l3 and w2<->w3
        mov l2, l3
        mov l3, w4

        mov w4, w2
        mov w2, w3
        mov w3, w4

    update_timeslot_table_compare3:
        cp l2, l1           ; l2 >= l1
        brsh update_timeslot_table_compare_done ; if yes, jump away

        mov w4, l1          ; if no, swap l1<->l2 and w1<->w2
        mov l1, l2
        mov l2, w4

        mov w4, w1
        mov w1, w2
        mov w2, w4

    update_timeslot_table_compare_done:
        ldi XL, low(PWMActionSlots) ; init X register to point to the ActionSlots table
        ldi XH, high(PWMActionSlots)

        ldi w6, 0           ; init persistent mask

        mov w4, l1          ; test if l1 is before OF1, if no, insert OF1
        rcall update_timeslot_test_of1
        cpse l1, l2         ; l1 == l2? if yes, skip next instruction
        rcall update_timeslot_write_timeslot_counter    ; search timeslot counter for this brightness value and write it

        or w1, w2           ; OR masks together
        mov w4, l2          ; test if l2 is before OF1, if no, insert OF1
        rcall update_timeslot_test_of1
        cpse l2, l3         ; l2 == l3? if yes, skip next instruction
        rcall update_timeslot_write_timeslot_counter    ; search timeslot counter for this brightness value and write it

        or w1, w3           ; OR masks together
        mov w4, l3          ; test if l3 is before OF1, if no, insert OF1
        rcall update_timeslot_test_of1
        mov w4, l3          ; copy w4 again, since update_timeslot_test_of1 overwrites it
        rcall update_timeslot_write_timeslot_counter    ; search timeslot counter for this brightness value and write it

        ldi w4, 0xff        ; insert OF1 if not already inserted...
        rcall update_timeslot_test_of1

        ldi w4, 0           ; store OF0
        st X+, w4
        st X+, w4
        ldi w4, 0          ; store bitmask to identify OF0
        st X+, w4


        sbr Flags, (1<<PWMOverflowSlot) ; next slot is burst
        ldi w1, 0
        mov PWMBitmask, w1

        ldi w1, low(PWMActionSlots)
        ldi w2, high(PWMActionSlots)
        sub XL, w1          ; calculate slot counter, result is in XL
        sbc XH, w2

        ldi w1, 0
    update_timeslot_table_calc_counter:  ; DIV for poor people...
        inc w1
        dec XL
        dec XL
        dec XL
        brne update_timeslot_table_calc_counter ; loop while XL nonzero, result is in w1

        sts PWMActionSlotCounter, w1    ; store slot counter

        ldi w1, low(PWMActionSlots) ; store pointer to the first timeslot
        sts PWMActionSlotPointer, w1
        ldi w1, high(PWMActionSlots)
        sts PWMActionSlotPointer+1, w1

        ret

; }}}

;   function: update_pwm_burst_table {{{
; =============================================================================
update_pwm_burst_table:
        ldi XL, low(PWMBurstOutputTable)
        ldi XH, high(PWMBurstOutputTable)

        ldi w1, 0                   ; init pulse counter                                            1
update_pwm_burst_table_loop:
        ldi w2, 0xff                ; initial output value (inverted)                               1
update_channel0:
        lds w3, LedStatusTable+LedRecSize*0+LedRecBrightness  ; load channel0 desired brightness value
        cp w1, w3                   ; channel 0 brightness reached, w1 >= ch0?                      1
        brsh update_channel1        ; if true, jump away                                            1 if false, 2 if true
        cbr w2, (1<<PWMChannel0)    ; if false, clear bit for channel0                              1
update_channel1:
        lds w3, LedStatusTable+LedRecSize*1+LedRecBrightness  ; load channel1 desired brightness value
        cp w1, w3                   ; channel 1 brightness reached, w1 >= ch1?                      1
        brsh update_channel2        ; if true, jump away                                            1 if false, 2 if true
        cbr w2, (1<<PWMChannel1)    ; if false, clear bit for channel1                              1
update_channel2:
        lds w3, LedStatusTable+LedRecSize*2+LedRecBrightness  ; load channel2 desired brightness value
        cp w1, w3                   ; channel 2 brightness reached, w1 >= ch2?                      1
        brsh update_store_output    ; if true, jump away                                            1 if false, 2 if true
        cbr w2, (1<<PWMChannel2)    ; if false, clear bit for channel2                              1
update_store_output:
        st X+, w2                   ; store inverted output value                                   2
        inc w1                      ;                                                               1
        cpi w1, PWMBurstLevels      ;                                                               1
        brne update_pwm_burst_table_loop ;                                                          1 if false, 2 if true
        ret                         ;                                                               4
        ; }}}

;   function: update_status_table {{{
; =============================================================================
update_status_table:
        ldi YH, high(LedStatusTable)                     ; load LedStatusTable address into Y
        ldi YL, low(LedStatusTable)
        ldi w1, 3
        mov l1, w1

us_update_channels:                                      ; update all three channels
        ; loop body: update current channel {{{
        ldd w1, Y+LedRecBrightness                  ; load channel brightness
        ldd w2, Y+LedRecTargetVal                   ; load channel target value
        cp w2, w1
        breq us_cleanup                             ; if equal, jump to cleanup, nothing to do
        ldd w3, Y+LedRecCycleCount                  ; load cycle counter

        ldd w4, Y+LedRecFadeSpeedL                  ; load fade speed, byte below 1
        ldd w5, Y+LedRecFadeSpeedH                  ; load fade speed, byte above 1
        add w3, w4                                  ; add low byte
        clr w4                                      ; clear w4
        adc w4, w5                                  ; add high byte, now w4 is brightstep.

        std Y+LedRecCycleCount, w3                  ; store low byte in cycle count for next round

        cp w2, w1
        brlo us_decrement                           ; if target val < brightness, jump to decrement part

        add w1, w4                                  ; increase brightness
        brcs us_target_missed                       ; if an overflow happened, just set the target value

        cp w2, w1                                   ; compare brightness after increase and target brigtness
        brsh us_store                               ; if current <= target, store brightness
        rjmp us_target_missed                       ; else we added too much, just set the target value...

us_decrement:
        sub w1, w4                                  ; decrease brightness
        brcs us_target_missed                       ; if an underflow happened, just set the target value

        cp w1, w2                                   ; compare current and target value
        brsh us_store                               ; if current >= target, store current brightness
                                                    ; else we missed the target

us_target_missed:
        mov w1, w2                                  ; set the target value

us_store:
        cp w1, w2                                   ; compare current brightness and target brightness
        brne us_store2                              ; if not equal, just store the new brightness value
        ldd w4, Y+LedRecFlags                       ; load led flags
        sbr w4, (1<<LedFlagTargetReached)           ; set the target reached flag
        std Y+LedRecFlags, w4                       ; store the new flags
us_store2:
        std Y+LedRecBrightness, w1                  ; store brightness

us_cleanup:
        ; }}}
        adiw YH:YL, LedRecSize
        dec l1
        brne us_update_channels
        ret

; }}}

;   function: check_serial_input {{{
; =============================================================================
check_serial_input:
        MACRO_FIFO_LOAD_COUNTER UReceiveFifo, w1   ; load uart receive fifo counter into w1
        cpi w1, 0
        breq check_ret1 ; if not, return

        MACRO_FIFO_READ UReceiveFifo, w1, URSize   ; else load received byte into w1

        ldi YH, high(LedStatusTable)                     ; load LedStatusTable address into Y
        ldi YL, low(LedStatusTable)

        rjmp check

check_ret1:
        ret

check:
        ldi w2, 1

        cpi w1, '1'
        brne check_4

        ldd w1, Y+LedRecTargetVal+0*LedRecSize
        sub w1, w2
        std Y+LedRecTargetVal+0*LedRecSize, w1
        rjmp check_end

check_4:
        cpi w1, '4'
        brne check_2

        ldd w1, Y+LedRecTargetVal+0*LedRecSize
        add w1, w2
        std Y+LedRecTargetVal+0*LedRecSize, w1
        rjmp check_end

check_2:
        cpi w1, '2'
        brne check_5

        ldd w1, Y+LedRecTargetVal+1*LedRecSize
        sub w1, w2
        std Y+LedRecTargetVal+1*LedRecSize, w1
        rjmp check_end

check_5:
        cpi w1, '5'
        brne check_3

        ldd w1, Y+LedRecTargetVal+1*LedRecSize
        add w1, w2
        std Y+LedRecTargetVal+1*LedRecSize, w1
        rjmp check_end

check_3:
        cpi w1, '3'
        brne check_6

        ldd w1, Y+LedRecTargetVal+2*LedRecSize
        sub w1, w2
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_6:
        cpi w1, '6'
        brne check_0

        ldd w1, Y+LedRecTargetVal+2*LedRecSize
        add w1, w2
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_0:
        cpi w1, '0'
        brne check_enter

        ldi w1, 0
        std Y+LedRecTargetVal+0*LedRecSize, w1
        std Y+LedRecTargetVal+1*LedRecSize, w1
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_enter:
        cpi w1, 13
        brne check_plus

        ldd w1, Y+LedRecBrightness+0*LedRecSize
        std Y+LedRecTargetVal+0*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+1*LedRecSize
        std Y+LedRecTargetVal+1*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+2*LedRecSize
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_plus:
        cpi w1, '+'
        brne check_minus

        ldd w1, Y+LedRecBrightness+0*LedRecSize
        inc w1
        std Y+LedRecTargetVal+0*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+1*LedRecSize
        inc w1
        std Y+LedRecTargetVal+1*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+2*LedRecSize
        inc w1
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_minus:
        cpi w1, '-'
        brne check_pgm

        ldd w1, Y+LedRecBrightness+0*LedRecSize
        dec w1
        std Y+LedRecTargetVal+0*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+1*LedRecSize
        dec w1
        std Y+LedRecTargetVal+1*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+2*LedRecSize
        dec w1
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_pgm:
        cpi w1, 'x'
        brne check_script
        rjmp 0xc00

check_script:
        cpi w1, 's'
        brne check_script_r

        ldi w2, (1<<ScriptEnabled)
        eor Flags, w2
        rjmp check_end

check_script_r:
        cpi w1, 'r'
        brne check_script_g

        lds w1, ScriptInfoTable+ScrRecHandler
        ldi w2, 0b10000000
        eor w1, w2
        sts ScriptInfoTable+ScrRecHandler, w1

        rjmp check_end

check_script_g:
        cpi w1, 'g'
        brne check_script_b

        lds w1, ScriptInfoTable+ScrRecSize+ScrRecHandler
        ldi w2, 0b10000000
        eor w1, w2
        sts ScriptInfoTable+ScrRecSize+ScrRecHandler, w1

        rjmp check_end

check_script_b:
        cpi w1, 'b'
        brne check_end

        lds w1, ScriptInfoTable+2*ScrRecSize+ScrRecHandler
        ldi w2, 0b10000000
        eor w1, w2
        sts ScriptInfoTable+2*ScrRecSize+ScrRecHandler, w1

        rjmp check_end


check_end:
        ret
; }}}

;   function: write_uart_transmit_fifo {{{
; =============================================================================
; data byte is loaded into w1
write_uart_transmit_fifo:
        MACRO_FIFO_LOAD_COUNTER UTransmitFifo, w2   ; load uart transmit fifo counter into w2
        cpi w2, UTSize ; compare counter and fifo size
        breq write_uart_discard_data ; if counter == fifosize, just discard the byte received

        movw l2:l1, YH:YL           ; save Y

        MACRO_FIFO_STORE UTransmitFifo, w1, w2, UTSize
        sbi UCSRB, UDRIE            ; enable data register empty interrupt
        movw YH:YL, l2:l1           ; restore Y

write_uart_discard_data:

        ret
; }}}


;   tables
; =============================================================================

;   data segment(ram), dynamic tables
; =============================================================================
.dseg

;   table: PWMBurstOutputTable {{{
; =============================================================================
; The burst table contains a bitmask for the LEDs for each pwm cycle in burst
; mode

PWMBurstOutputTable:
        .byte   PWMBurstLevels+1

; }}}

;   table: LedStatusTable {{{
; =============================================================================
; The status table contains a record for each LED.
; This record consists of
;   .byte   : (LedRecBrightness) Current brightness value
;   .byte   : (LedRecTargetVal)  Target brightness value (0-255)
;   .byte   : (LedRecFadeSpeedL) 8.8 fixed point value
;   .byte   : (LedRecFadeSpeedH)
;   .byte   : (LedRecCycleCount) Cycle counter
;   .byte   : (LedRecFlags)      Flags: TargetReached, ...

LedStatusTable:
        .byte   PWMChannels * LedRecSize

; }}}

;   table: PWMActionSlots {{{
; =============================================================================
; the action slot table contains the next timeslot (2 bytes) and a
; bitmask (1 byte), which is XORed to the actual output register, for
; clearing bytes
;
; by using only 3 channels, 5 slots are needed at max
;
;   .byte   : (ActionRecTimeSlotL)  timer counter value
;   .byte   : (ActionRecTimeSlotH)  "
;   .byte   : (ActionRecBitmask)    XOR bitmask

PWMActionSlots:
        .byte   (PWMChannels+2) * 3
PWMActionSlotPointer:
        .byte   2
PWMActionSlotCounter:
        .byte   1

; }}}

;   program segment(flash), static tables
; =============================================================================
.cseg

;   table: PWMTimerTable {{{
; =============================================================================
; timer1 TOP values
PWMTimerTable:
    ; first 16 slots are cumulatede into one cumulated into one burst
    ;.dw 0, 2, 8, 18, 31, 49, 71, 96
    ;.dw 126, 159, 197, 238, 283, 333, 386, 443
    .dw 504, 569, 638, 711, 787, 868, 953, 1041
    .dw 1134, 1230, 1331, 1435, 1543, 1655, 1772, 1892
    .dw 2016, 2144, 2276, 2411, 2551, 2695, 2842, 2994
    .dw 3150, 3309, 3472, 3640, 3811, 3986, 4165, 4348
    .dw 4535, 4726, 4921, 5120, 5323, 5529, 5740, 5955
    .dw 6173, 6396, 6622, 6852, 7087, 7325, 7567, 7813
    .dw 8063, 8317, 8575, 8836, 9102, 9372, 9646, 9923
    .dw 10205, 10490, 10779, 11073, 11370, 11671, 11976, 12285
    .dw 12598, 12915, 13236, 13561, 13890, 14222, 14559, 14899
    .dw 15244, 15592, 15945, 16301, 16661, 17025, 17393, 17765
    .dw 18141, 18521, 18905, 19293, 19685, 20080, 20480, 20884
    .dw 21291, 21702, 22118, 22537, 22960, 23387, 23819, 24254
    .dw 24693, 25135, 25582, 26033, 26488, 26946, 27409, 27876
    .dw 28346, 28820, 29299, 29781, 30267, 30757, 31251, 31750
    .dw 32251, 32757, 33267, 33781, 34299, 34820, 35346, 35875
    .dw 36409, 36946, 37488, 38033, 38582, 39135, 39692, 40253
    .dw 40818, 41387, 41960, 42537, 43117, 43702, 44291, 44883
    .dw 45480, 46080, 46684, 47293, 47905, 48521, 49141, 49765
    .dw 50393, 51025, 51661, 52300, 52944, 53592, 54243, 54899
    .dw 55558, 56222, 56889, 57560, 58235, 58914, 59598, 60285
    .dw 60975, 61670, 62369, 63072, 63779, 489, 1204, 1922
    .dw 2645, 3371, 4101, 4836, 5574, 6316, 7062, 7812
    .dw 8566, 9324, 10085, 10851, 11621, 12394, 13172, 13954
    .dw 14739, 15528, 16322, 17119, 17920, 18725, 19534, 20347
    .dw 21164, 21985, 22810, 23638, 24471, 25308, 26148, 26993
    .dw 27841, 28693, 29550, 30410, 31274, 32142, 33014, 33890
    .dw 34770, 35654, 36542, 37433, 38329, 39229, 40132, 41040
    .dw 41951, 42866, 43786, 44709, 45636, 46567, 47502, 48441
    .dw 49384, 50331, 51282, 52236, 53195, 54158, 55124, 56095
    .dw 57069, 58047, 59030, 60016, 61006, 62000, 62998, 0


; }}}


;   static scripts
; =============================================================================

.include "script_macros.inc"
.include "static_scripts.inc"
